// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//						(c) 1997 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	epfast.cc --- Count rate dependent gain correction task
//
//	DESCRIPTION:
//		This is the task to perform the count rate dependent gain
//              correction for PN Timing and Burst mode data
//
//	CHANGE HISTORY:
//		RDS, 2007-09-26: 1.0
//
//	AUTHOR:
//		R. D. Saxton
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.
//		[2]: Barycen task description, U. Lammers (SOC), G. Vacanti (SOC)
//
#include "param.h"
#include "errstr.h"
#include "Tracer.h"
#include "OalServer.h"
#include "Dal.h"
#include "epfast.h"
#include "Xmm.h"
#include "CalServer.h"
#include "EnergyCorrector.h"
//gcc4.8.2 portability change
#include <unistd.h>

//using namespace std;

void epfast()
{
   PROC

        // Set the instrument in the CAL
        calServer.setInstrument(Xmm::EPN);

	//
	//	Open the event file
	//
        string evfile = stringParameter("evfile");
        DataSet * evset = dataSetServer->open(evfile, DataSet::Modify);

        //Table *table = evset->table("EVENTS", Dal::Modify);
        Table *table = evset->table("EVENTS", Dal::Modify);

        /* If the offset table has been calculated with the closed filter 
           this routine should not be executed */
        if (table->hasAttribute("OTFILTER")) { 
           if (trim(uppercase(table->attribute("OTFILTER")->asString())) 
                                                         == "CLOSED") {
              errstr<<"epfast will not run on data processed with an "
                    <<"offset map taken with the closed filter"
                    <<Sas::error("OTFCLOSED");
           }
        }

	// Check if this is an EPN observation
	if (table->attribute("INSTRUME")->asString() != "EPN") {
		errstr<<table->qualifiedName()<<" is not an Epic-PN data set - "
					"epfast conversion is not possible"
				  <<Sas::error("ThisIsNotEPN");
        }

        // Don't apply if the RDPHA correction has been applied
        if (table->hasAttribute("PHA_RDCO")) { 
           if (table->attribute("PHA_RDCO")->asBool()) {
              errstr<<"epfast will not run on data which has been "
                    <<"corrected with the RDPHA correction"
                    <<Sas::error("RDPHAcorrected");
           }
        }

        // Don't apply if this RDCTI correction has already been applied
        if (table->hasAttribute("RDCTI")) { 
           if (table->attribute("RDCTI")->asBool()) {
              errstr<<"The rate-dependent CTI correction has already "
                    <<"been applied to this eventlist. epfast will not "
                    <<"be run again."<<Sas::error("RDCTIcorrected");
           }
        }

	// Check if this is an fast mode observation and set the 
        // upper bound in RAW-Y for the event extraction.
        int rawy_upper=0;
        double expfrac=0.0;

        if (table->attribute("SUBMODE")->asString() == "FastTiming") {
           calServer.state()->mode()->set(Xmm::FAST_TIMING);
           rawy_upper = 200;
           expfrac = 0.99117;
        } else if (table->attribute("SUBMODE")->asString() == "FastBurst") {
           calServer.state()->mode()->set(Xmm::FAST_BURST);
           rawy_upper = 140;
           expfrac = 0.02983;
        } else { 
           errstr<<table->qualifiedName()<<" is not a fast mode data "
                         "set - epfast conversion is not possible"
                                  <<Sas::error("ThisIsNotFAST");
        }

        //cout<<"RAWY range 0 - "<<rawy_upper<<endl;

       /* Set the instrument */
        const EPNcam _ePNCam;

       /* Get parameters from environment */
        double tbinwidth = realParameter("tbinwidth");
        int pimin = intParameter("pimin");
        int pimax = intParameter("pimax");

       /* Get the times, energies and RAWX values of each event */
        Column *tcol = table->column("TIME"); 
        Column *pcol = table->column("PI"); 
        Column *ptcol = table->column("PATTERN"); 
        Column *fcol = table->column("FLAG"); 
        Column *rxcol = table->column("RAWX"); 
        Column *rycol = table->column("RAWY"); 
        ColumnData *cdtime = tcol->data(),
                   *cdPI = pcol->data(),
                   *cdPATT = ptcol->data(),
                   *cdF = fcol->data(),
                   *cdX = rxcol->data(),
                   *cdY = rycol->data();
        real64  *time = cdtime->real64Data();
        int16 *rawx = cdX->int16Data(),
              *rawy = cdY->int16Data();
        int8  *pattern = cdPATT->int8Data();
        int32 *flag = cdF->int32Data();
      
       // Set the number of rows
	unsigned N = table->rows();

        vector<int32> pichan(N,0);
        int32 *pichan32 = 0;
        int16 *pichan16 = 0;
        if (pcol->dataType()==Dal::Int32) {
          pichan32 = cdPI->int32Data();
          for (unsigned i=0; i<N; ++i)
                        pichan[i] = pichan32[i];
        } else {
          //pichan16 = (int32*)cdPI->int16Data();
          pichan16 = cdPI->int16Data();
          for (unsigned i=0; i<N; ++i)
                pichan[i] = static_cast<int32>(pichan16[i]);
        }

       /*
       // Sort out the datatype of the PI column 
        if (pcol->dataType()==Dal::Int32) {
      cout<<"INT32"<<endl;
           int32 *pichan = cdPI->int32Data();
        } 
        else { // INT16 
      cout<<"INT16"<<endl;
           int32 *pichan = cdPI->int16Data();
           for (unsigned i=0; i<N; ++i)
               pichan[0] = static_cast<int32>(pichan[i]);
        }
       */

       /* Find the extremes of the times from these events */
        double tmin=999999999999999.0;
        double tmax=0.0;
        for (unsigned i=0; i<N; ++i)
        {
            if (time[i]<tmin) {tmin=time[i];}
            if (time[i]>tmax) {tmax=time[i];}
        }
        unsigned int Ntimebins = INT( (tmax-tmin)/tbinwidth ) + 1;
        //cout<<"Time range: "<<tmin<<" "<<tmax<<" "<<Ntimebins<<" "<<N<<endl;
            
      /* Get the conversion factors from the CCF */
        EnergyCorrector *gainCorrector = 0;
        gainCorrector = calServer.getAtom(gainCorrector);
        vector<float> rdc = gainCorrector->getRateDepGainCoeffs(); 
        float a0 = rdc[0], a1 = rdc[1], a2 = rdc[2];
 
        errstr << "Using Coefficients: "<<a0<<" "<<a1<<" "<<a2
         << Sas::message(Msg::App, Msg::Verbose);

        calServer.releaseAtom(gainCorrector);
       
      // Set a vector for the time bin widths and initialise 
        vector<float> exposure(Ntimebins+1,0.0);

      /* Get the GTIs */
        if (evset->hasBlock("STDGTI04"))
        {
           Table *gtiTab = evset->table("STDGTI04");
           const Column *gtStartCol = gtiTab->column("START",Dal::Read);
           ColumnData *cGtiStartData = gtStartCol->data();
           real64 *gtiStart = cGtiStartData->real64Data();
           const Column *gtStopCol = gtiTab->column("STOP",Dal::Read);
           ColumnData *cGtiStopData = gtStopCol->data();
           real64 *gtiStop = cGtiStopData->real64Data();

         // Use the GTIs to find the exposure time in each bin
           for (unsigned t=1; t<=Ntimebins; ++t)
           {
             double tlower = tmin + (t-1) * tbinwidth;
             double tupper = min(tmax,(tmin + t*tbinwidth));

            // Loop over the GTIs
             for (unsigned n=0; n<gtiTab->rows(); ++n)
             {
               if (gtiStart[n] < tupper && gtiStop[n] > tlower)
                          exposure[t] += min(tupper, gtiStop[n]) - 
                                          max(tlower, gtiStart[n]);
             }
           }

         // Clean up
           delete cGtiStartData;
           delete cGtiStopData;

        } else {

          // Set the timebins conventionally
           for (unsigned t=1; t<=Ntimebins; ++t)
              exposure[t] = tbinwidth;

          // Make sure the last bin has the correct exposure
           exposure[Ntimebins] = tmax - (tmin + (Ntimebins-1) * tbinwidth);
        }

      /* Create an empty time series */
	vector<vector<int> > sumPI(_ePNCam.maxX()+1, 
                             vector<int>(Ntimebins+1,0));
	vector<vector<float> > ratePI(_ePNCam.maxX()+1, 
                             vector<float>(Ntimebins+1,0));

      /* Loop over events and construct time series containing the sum
	    of the PIs in that time bin */
        for (unsigned i=0; i<N; ++i)
        {
            // Use event if the column is correct, RAWY within correct range 
            // ,energy is within selected range flag=0 and pattern=0-4.
            // NB: The correction coeffs will ONLY work correctly
            // if the event file contains at least these selections. A
            // subset of pattern=0 for example will invalidate the result.
               if (pichan[i]<=pimax && pichan[i]>=pimin &&
                 rawx[i]<=INT(_ePNCam.maxX()) && rawx[i]>=INT(_ePNCam.minX())
                 && rawy[i]<=rawy_upper && pattern[i]<=4 && flag[i]==0)
	       {
                 // Calculate the time bin for this event 
	          unsigned tbin = INT(( time[i] - tmin ) / tbinwidth) + 1;

                 // Add the PI of this event into the time series
                  sumPI[rawx[i]][tbin] += pichan[i];
	       }
        }

      /* Convert time bins to rates in electrons per pixel per second */
      //  For each COLUMN
        for (unsigned col=_ePNCam.minX(); col<=_ePNCam.maxX(); ++col)
        {
           float meanrate = 0; 
           double sumsum = 0; 

          // For each TIMEBIN
           for (unsigned t=1; t<=Ntimebins; ++t)
           {
            // Change rawy_upper to raw-y range if lower becomes non-zero
             ratePI[col][t] = sumPI[col][t] / expfrac / 
                    		   exposure[t] / rawy_upper / 3.6;

           //cout<<"BIT:"<<col<<" rate: "<<ratePI[col][t]<<" sum:"<<sumPI[col][t]
             //  <<" expf: "<<expfrac<<" twid: "<<exposure[t]<<" rup:"
               //<<rawy_upper<<endl;

             meanrate += ratePI[col][t];
             sumsum += sumPI[col][t];
           }
          // Temp debug for Marcus - REMOVE !!!
           float corrfact = a0 * pow(meanrate,a1) + a2;

           errstr << "Mean rate for column:"<<col<<" = "
                  <<meanrate/Ntimebins<<"SUMPI:"<<sumsum<<" "
                  <<"Corrfact: "<<corrfact
                  << Sas::message(Msg::App, Msg::Noisy);
        }

       // Set the random seed
        srand(getpid());

       /* Apply correction to the PIs of each event */
	for (unsigned i=0; i<N; ++i)
        {
            // Find the time bin for this event
             unsigned tbin = INT(( time[i] - tmin ) / tbinwidth) + 1;

            // Find the gain correction factor given the PI rate in 
            // this time bin
            //int fred = pichan[i];
             if (rawx[i]<=INT(_ePNCam.maxX()) && 
                          rawx[i]>=INT(_ePNCam.minX()))
             {
                float corrfact = a0 * pow(ratePI[rawx[i]][tbin],a1) + a2;
                //cout<<"Fact: "<<ratePI[rawx[i]][tbin]<<" "<<corrfact<<endl;

              // Do the correction
              // First add on a random value between -0.5 and + 0.5
              // to avoid quantisation effects
                float ranbit = ( rand() % 100 ) // Gives a number from 0-9
                                 / 100.0 - 0.5;
                float pichanrand = pichan[i] + ranbit;

                if ( ! booleanParameter("withrandomisation")) {
                   pichanrand = float(pichan[i]);
                }

                if (pcol->dataType()==Dal::Int32) 
                   pichan32[i] = INT(pichanrand/corrfact +0.5);
                else
                   pichan16[i] = SHORT(pichanrand/corrfact +0.5);

                //cout<<"OUT: "<<pichan32[i]<<" "<<pichanrand<<" "<<corrfact<<endl; 
             }
        }
        
       /* Set a keyword to say that this file has been corrected */
        * table->addAttribute("RDCTI",
                  "Rate-dependent CTI correction applied ?") = true;

       /* Lets close up and go home */
        delete cdtime;
        delete cdPI;
        delete cdPATT;
        delete cdF;
        delete cdX;
        delete cdY;
        dataSetServer->close(evset);

   ENDPROC

} // epfast()

